@@ -26,13 +26,10 @@ use tokio_stream::wrappers::ReceiverStream;
 use uuid::Uuid;
 
 use super::{
-    CharacteristicFlags, CharacteristicReader, CharacteristicWriter, DescriptorFlags, WriteOp,
+    make_socket_pair, CharacteristicFlags, CharacteristicReader, CharacteristicWriter, DescriptorFlags, WriteOp,
     CHARACTERISTIC_INTERFACE, DESCRIPTOR_INTERFACE, SERVICE_INTERFACE,
 };
-use crate::{
-    make_socket_pair, parent_path, Adapter, Error, ErrorKind, Result, SessionInner, ERR_PREFIX, SERVICE_NAME,
-    TIMEOUT,
-};
+use crate::{parent_path, Adapter, Error, ErrorKind, Result, SessionInner, ERR_PREFIX, SERVICE_NAME, TIMEOUT};
 
 pub(crate) const MANAGER_INTERFACE: &str = "org.bluez.GattManager1";
 
@@ -312,6 +309,7 @@ pub enum CharacteristicWriteMethod {
     /// Call specified function for each write request.
     Fun(CharacteristicWriteFun),
     /// Provide written data over asynchronous IO functions.
+    /// This has low overhead.
     ///
     /// Use [CharacteristicControl] to obtain reader.
     Io,
@@ -377,6 +375,7 @@ pub enum CharacteristicNotifyMethod {
     /// Call specified function when client starts a notification session.
     Fun(CharacteristicNotifyFun),
     /// Write notify data over asynchronous IO.
+    /// This has low overhead.
     ///
     /// Use [CharacteristicControl] to obtain writer.
     Io,
@@ -606,9 +605,9 @@ impl CharacteristicWriteIoRequest {
     /// Accept the write request.
     pub fn accept(self) -> Result<CharacteristicReader> {
         let CharacteristicWriteIoRequest { mtu, tx, .. } = self;
-        let (fd, stream) = make_socket_pair()?;
+        let (fd, stream) = make_socket_pair(false)?;
         let _ = tx.send(Ok(fd));
-        Ok(CharacteristicReader { mtu: mtu.into(), stream })
+        Ok(CharacteristicReader { mtu: mtu.into(), stream, buf: Vec::new() })
     }
 
     /// Reject the write request.
@@ -915,8 +914,10 @@ impl RegisteredCharacteristic {
                             Some(CharacteristicNotify { method: CharacteristicNotifyMethod::Io, .. }) => {
                                 // BlueZ has already confirmed the start of the notification session.
                                 // So there is no point in making this fail-able by our users.
-                                let (fd, stream) = make_socket_pair().map_err(|_| ReqError::Failed)?;
-                                let writer = CharacteristicWriter { mtu: options.mtu.into(), stream };
+                                let (fd, stream) = make_socket_pair(true).map_err(|_| ReqError::Failed)?;
+                                // WORKAROUND: BlueZ drops data at end of packet if full MTU is used.
+                                let mtu = options.mtu.saturating_sub(5).into();
+                                let writer = CharacteristicWriter { mtu, stream };
                                 let _ = reg
                                     .c
                                     .control_handle
